package org.objectstyle.dataService;

import java.util.List;

import org.objectstyle.cayenne.access.DataContext;
import org.objectstyle.cayenne.exp.Expression;
import org.objectstyle.cayenne.query.Ordering;
import org.objectstyle.cayenne.query.SelectQuery;
import org.objectstyle.cayenne.xml.XMLEncoder;

/**
 * Provides a thin wrapper around the Cayenne DataContext and XMLEncoder service,
 * to return query results in XML, suitable for binding to a browser through Ajax.
 * 
 * @author cris.daniluk
 */
public class GenericDataService {

    protected DataContext dataContext;
    protected XMLEncoder xmlEncoder;
    
    /**
     * Initializes the DataService and binds it to the DataContext provided.
     * @param context
     */
    public GenericDataService(DataContext dataContext) {
        this.dataContext = dataContext;
        xmlEncoder = new XMLEncoder();
    }
    
    public GenericDataService(DataContext dataContext, String mappingFile) {
        this.dataContext = dataContext;
        xmlEncoder = new XMLEncoder(mappingFile);
    }

    /**
     * Runs a query for all objects of the specified <code>objectType</code>.
     * Can optionally be filtered by an <code>expression</code> and ordered by
     * the <code>sortList</code>. Use <code>filterInMemory</code> to determine
     * whether or not expressions and orderings should be applied at the database
     * or in memory.
     * @param objectType
     * @param expression
     * @param sortList
     * @param filterInMemory
     */
    public String returnObjects(Class objectType, Expression expression, List sortList, boolean filterInMemory) {
        
        SelectQuery query;
        if (expression != null && !filterInMemory) {
            query = new SelectQuery(objectType, expression);
            
            if (sortList != null) {
                query.addOrderings(sortList);
            }
        } else {
            query = new SelectQuery(objectType);
        }
        
        List queryResults = dataContext.performQuery(query);
        
        if (filterInMemory) {
            
            if (expression != null) {
                queryResults = expression.filterObjects(queryResults);
            }
            
            if (sortList != null) {
                Ordering.orderList(queryResults, sortList);
            }
        }
        
        return xmlEncoder.encode("resultSet", queryResults);
    }

    /**
     * Runs the named query <code>queryName</code>. Can optionally be
     * filtered by an <code>expression</code> and ordered by the
     * <code>sortList</code>. Use <code>filterInMemory</code> to determine
     * whether or not expressions and orderings should be applied at the
     * database or in memory.
     * 
     * @param queryName
     * @param expression
     * @param sortList
     * @param filterInMemory
     */
    public String returnStoredQuery(String queryName, Expression expression, List sortList, boolean filterInMemory) {
        SelectQuery query;
        
        if (filterInMemory) {
            // warn user
        }
        
        query = (SelectQuery) dataContext.getEntityResolver().getQuery(queryName);
        
        if (!filterInMemory && sortList != null) { 
            query.addOrderings(sortList);
        }
        
        List queryResults = dataContext.performQuery(query);
        
        if (filterInMemory) {
            
            if (expression != null) {
                queryResults = expression.filterObjects(queryResults);
            }
            
            if (sortList != null) {
                Ordering.orderList(queryResults, sortList);
            }
        }
        
        return xmlEncoder.encode("resultSet", queryResults);
    }

}
