package org.objectstyle.dataService.servlet;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang.StringUtils;
import org.objectstyle.cayenne.access.DataContext;
import org.objectstyle.cayenne.exp.Expression;
import org.objectstyle.cayenne.query.Ordering;
import org.objectstyle.dataService.GenericDataService;

/**
 * Creates a generic data requesting service to bind Cayenne objects directly to
 * the browser through Ajax.
 * 
 * Request parameters are:
 * 
 * <ul>
 * <li>objectType</li>
 * <li>query</li>
 * <li>expression</li>
 * <li>filterInMemory</li>
 * <li>ordering</li>
 * </ul>
 * 
 * @author cris.daniluk@gmail.com
 */
public abstract class AbstractDataServiceServlet extends HttpServlet {

    /**
     * Request parameter <code>objectType</code>, which should correspond to
     * a key in the map returned by <see>getObjectMap</code>.
     * 
     * Note that this parameter is mutually exclusive with query. Both should
     * not be used.
     */
    protected static final String OBJECT_TYPE = "objectType";

    /**
     * Request parameter <code>expression</code>, which specifies an optional
     * expression to be run with the query.
     */
    protected static final String EXPRESSION = "expression";

    /**
     * Request parameter <code>query</code>, which specifies a named Cayenne
     * query that should be used to retrieve results. Note that the data service
     * will not alter the refetch or caching properties of the query.
     * 
     * Note that this parameter is mutually exclusive with objectType. Both
     * should not be used.
     */
    protected static final String QUERY = "query";

    /**
     * Request parameter <code>filterInMemory</code>, which specifies whether
     * or not the expression (if provided) should be executed in memory or at
     * the database. By default, this will be true if the query is named (to
     * ensure caching), or false if the query is not named.
     */
    protected static final String FILTER_IN_MEMORY = "filterInMemory";

    /**
     * Request parameter <code>ordering</code>, which specifies the sort
     * ordering of results. If filterInMemory is set to true, the results will
     * also be ordered in memory. The sort order will match the query or
     * database default if not specified. The ordering should be provided as a
     * comma separated list of field names (from the ObjEntity, not from the
     * DbEntity). To specify ascending or descending, prefix the field name with
     * a + or -. The default order is ascending.
     */
    protected static final String ORDERING = "ordering";

    protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {

        DataContext context = DataContext.getThreadDataContext();

        GenericDataService dataService = null;
        if (getMappingFile() != null) {
            dataService = new GenericDataService(context, getMappingFile());
        } else {
            dataService = new GenericDataService(context);
        }

        String objectTypeName = request.getParameter(OBJECT_TYPE);
        String queryName = request.getParameter(QUERY);

        String expressionName = request.getParameter(EXPRESSION);

        String filterInMemoryValue = request.getParameter(FILTER_IN_MEMORY);

        Expression expression = null;
        if (StringUtils.isNotBlank(expressionName)) {

            Map expressionMap = getExpressionMap();

            if (expressionMap == null) {
                expression = Expression.fromString(expressionName);
            } else if (expressionMap.size() == 1) {
                expression = null;
            } else {
                expression = (Expression) expressionMap.get(expressionName);
            }
        }

        String orderingValue = request.getParameter(ORDERING);
        
        List sortList = null;
        if (StringUtils.isNotBlank(orderingValue)) {
            
            sortList = new ArrayList();

            String[] orderings = orderingValue.split(",");

            for (int i = 0; i < orderings.length; i++) {

                boolean ascending = true;
                char firstChar = orderings[i].charAt(0);

                if (firstChar == '-') {
                    orderings[i] = orderings[i].substring(1);
                    ascending = false;
                } else {
                    orderings[i] = orderings[i].substring(1);
                }

                sortList.add(new Ordering(orderings[i], ascending));
            }
        }

        boolean filterInMemory = true;
        if (StringUtils.isNotBlank(filterInMemoryValue)) {
            filterInMemory = Boolean.valueOf(filterInMemoryValue).booleanValue();
        }
        
        String resultSet = null;
        if (StringUtils.isNotBlank(objectTypeName)) {

            resultSet = dataService.returnObjects((Class) getObjectMap().get(objectTypeName), expression, sortList,
                    filterInMemory);
        } else if (StringUtils.isNotBlank(queryName)) {

            resultSet = dataService.returnStoredQuery(queryName, expression, sortList, filterInMemory);

        } else {
            // Invalid query
            throw new ServletException("Invalid query type");
        }

        response.setContentType("application/xml");
        PrintWriter writer = response.getWriter();
        writer.print(resultSet);
        writer.flush();
    }

    protected void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException,
            IOException {
        doGet(request, response);
    }

    /**
     * Provides a mapping of objects to Cayenne objects. This ensures that data
     * objects that are not intended to be directly exposed to the browser
     * and/or user cannot be accessed. This method must be overridden.
     * 
     * @return
     */
    protected abstract Map getObjectMap();
    
    /**
     * This method can be overridden to specify an XML mapping file. By default, no mapping file is used.
     * @return null
     */
    protected String getMappingFile() {
        return null;
    }

    /**
     * This method can be overridden to validate filters, providing a mechanism
     * to prevent potentially abusive queries, etc.
     * 
     * @see org.objectstyle.cayenne.exp.Expression
     * @return
     */
    protected boolean expressionValidator() {
        return true;
    }

    /**
     * This method can be overriden to provide a list of expressions that may be
     * run. If you are concerned about the security of allowing users to execute
     * arbitrary expressions, use this method. To disable support for
     * expressions, you can override this method and return an empty map. If
     * this method is not extended, all expressions will be permitted.
     * 
     * @see org.objectstyle.cayenne.exp.Expression
     * @return Map of Strings to Expressions
     */
    protected Map getExpressionMap() {
        return Collections.EMPTY_MAP;
    }

}
