package org.objectstyle.dataService.webwork;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang.StringUtils;
import org.objectstyle.cayenne.access.DataContext;
import org.objectstyle.cayenne.exp.Expression;
import org.objectstyle.cayenne.query.Ordering;
import org.objectstyle.dataService.GenericDataService;

import com.opensymphony.webwork.interceptor.ServletResponseAware;
import com.opensymphony.xwork.ActionSupport;

/**
 * @author cris.daniluk@gmail.com
 */
public abstract class AbstractDataServiceAction extends ActionSupport implements ServletResponseAware {
    
    protected HttpServletResponse response;

    /**
     * Request parameter <code>objectType</code>, which should correspond to
     * a key in the map returned by <see>getObjectMap</code>.
     * 
     * Note that this parameter is mutually exclusive with query. Both should
     * not be used.
     */
    protected String objectType;

    /**
     * Request parameter <code>expression</code>, which specifies an optional
     * expression to be run with the query.
     */
    protected String expression;

    /**
     * Request parameter <code>query</code>, which specifies a named Cayenne
     * query that should be used to retrieve results. Note that the data service
     * will not alter the refetch or caching properties of the query.
     * 
     * Note that this parameter is mutually exclusive with objectType. Both
     * should not be used.
     */
    protected String query;

    /**
     * Request parameter <code>filterInMemory</code>, which specifies whether
     * or not the expression (if provided) should be executed in memory or at
     * the database. By default, this will be true if the query is named (to
     * ensure caching), or false if the query is not named.
     */
    protected Boolean filterInMemory;

    /**
     * Request parameter <code>ordering</code>, which specifies the sort
     * ordering of results. If filterInMemory is set to true, the results will
     * also be ordered in memory. The sort order will match the query or
     * database default if not specified. The ordering should be provided as a
     * comma separated list of field names (from the ObjEntity, not from the
     * DbEntity). To specify ascending or descending, prefix the field name with
     * a + or -. The default order is ascending.
     */
    protected String ordering;
    
    public String execute() throws Exception {
        
        DataContext context = DataContext.getThreadDataContext();
        
        GenericDataService dataService = null;
        if (getMappingFile() != null) {
            dataService = new GenericDataService(context, getMappingFile());
        } else {
            dataService = new GenericDataService(context);
        }

        Expression dataExpression = null;
        if (StringUtils.isNotBlank(expression)) {

            Map expressionMap = getExpressionMap();

            if (expressionMap == null) {
                dataExpression = Expression.fromString(expression);
            } else if (expressionMap.size() == 1) {
                dataExpression = null;
            } else {
                dataExpression = (Expression) expressionMap.get(expression);
            }
        }

        List sortList = null;
        if (StringUtils.isNotBlank(ordering)) {
            
            sortList = new ArrayList();

            String[] orderings = ordering.split(",");

            for (int i = 0; i < orderings.length; i++) {

                boolean ascending = true;
                char firstChar = orderings[i].charAt(0);

                if (firstChar == '-') {
                    orderings[i] = orderings[i].substring(1);
                    ascending = false;
                } else {
                    orderings[i] = orderings[i].substring(1);
                }

                sortList.add(new Ordering(orderings[i], ascending));
            }
        }

        String resultSet = null;
        if (filterInMemory == null) {
            filterInMemory = Boolean.TRUE;
        }
        
        if (StringUtils.isNotBlank(objectType)) {

            resultSet = dataService.returnObjects((Class) getObjectMap().get(objectType), dataExpression, sortList,
                    filterInMemory.booleanValue());
        } else if (StringUtils.isNotBlank(query)) {

            resultSet = dataService.returnStoredQuery(query, dataExpression, sortList, filterInMemory.booleanValue());

        } else {
            // Invalid query
            throw new Exception("Invalid query type");
        }
        
        response.setContentType("application/xml");
        PrintWriter out = response.getWriter();
        out.print(resultSet);
        
        return null;
    }
    
    /**
     * Provides a mapping of objects to Cayenne objects. This ensures that data
     * objects that are not intended to be directly exposed to the browser
     * and/or user cannot be accessed. This method must be overridden.
     * 
     * @return
     */
    protected abstract Map getObjectMap();

    /**
     * This method can be overridden to validate filters, providing a mechanism
     * to prevent potentially abusive queries, etc.
     * 
     * @see org.objectstyle.cayenne.exp.Expression
     * @return
     */
    protected boolean expressionValidator() {
        return true;
    }

    /**
     * This method can be overriden to provide a list of expressions that may be
     * run. If you are concerned about the security of allowing users to execute
     * arbitrary expressions, use this method. To disable support for
     * expressions, you can override this method and return an empty map. If
     * this method is not extended, all expressions will be permitted.
     * 
     * @see org.objectstyle.cayenne.exp.Expression
     * @return Map of Strings to Expressions
     */
    protected Map getExpressionMap() {
        return Collections.EMPTY_MAP;
    }
    
    /**
     * This method can be overridden to specify an XML mapping file. By default, no mapping file is used.
     * @return null
     */
    protected String getMappingFile() {
        return null;
    }

    public void setExpression(String expression) {
        this.expression = expression;
    }

    public void setFilterInMemory(Boolean filterInMemory) {
        this.filterInMemory = filterInMemory;
    }

    public void setObjectType(String objectType) {
        this.objectType = objectType;
    }

    public void setOrdering(String ordering) {
        this.ordering = ordering;
    }

    public void setQuery(String query) {
        this.query = query;
    }

    public void setServletResponse(HttpServletResponse response) {
        this.response = response;
    }
    
    

}
